# /psqtraviscontainer/container.py
#
# Abstract base class for an operating system container.
#
# See /LICENCE.md for Copyright information
"""Abstract base class for an operating system container."""

import abc

import os  # suppress(PYC50)

import re

import shutil

import subprocess

import sys

from collections import namedtuple

from contextlib import contextmanager

import parseshebang

import shutilwhich  # suppress(F401,PYC50,unused-import)

import six


@contextmanager
def updated_environ(prepend, overwrite):
    """Context with prepend added to and overwrite replacing os.environ."""
    env = os.environ.copy()
    for key, value in prepend.items():
        env[key] = "{0}{1}{2}".format(value,
                                      os.pathsep,
                                      env.get(key, ""))

    env.update(overwrite)

    old_environ = os.environ
    os.environ = env

    try:
        yield env
    finally:
        os.environ = old_environ


class AbstractContainer(six.with_metaclass(abc.ABCMeta, object)):
    """An abstract class representing an OS container."""

    PopenArguments = namedtuple("PopenArguments", "argv prepend overwrite")

    # vulture doesn't know that the __defaults__ attribute is actually
    # built-in.
    #
    # suppress(unused-attribute)
    PopenArguments.__new__.__defaults__ = (None, dict(), dict())

    @staticmethod
    def rmtree(directory):
        """Remove directory, but ignore errors."""
        try:
            shutil.rmtree(directory)
        except (shutil.Error, OSError):   # suppress(pointless-except)
            pass

    @abc.abstractmethod
    def _subprocess_popen_arguments(self, argv, **kwargs):
        """Return a PopenArguments tuple.

        This indicates what should be passed to subprocess.Popen when the
        execute method is called on this class.
        """
        del argv
        del kwargs

        raise NotImplementedError()

    @abc.abstractmethod
    def _package_system(self):
        """Return the package system this container should be using."""
        raise NotImplementedError()

    @abc.abstractmethod
    def clean(self):
        """Clean this container to prepare it for caching.

        Remove any non-useful files here.
        """
        raise NotImplementedError()

    def __enter__(self):
        """Use this container as a context."""
        return self

    def __exit__(self, exc_type, value, traceback):
        """Clean this container once it has been used a context."""
        del exc_type
        del value
        del traceback

        self.clean()

    def root_filesystem_directory(self):
        """Return absolute and real path to installed packages."""
        return os.path.realpath(self._root_filesystem_directory())

    def install_packages(self, repositories_path, packages_path):
        """Install packages and set up repositories as configured.

        :repositories_path: should be a path to a text file containing
                            a list of repositories to add to the package system
                            before installing any packages.
        :packages_path: should be a path to a text file containing a
                        list of packages to be installed.
        """
        if packages_path:
            package_system = self._package_system()

            # Add any repositories to the package system now
            if repositories_path:
                with open(repositories_path, "r") as repositories_file:
                    repo_lines = repositories_file.read().splitlines(False)

                package_system.add_repositories(repo_lines)

            with open(packages_path) as packages_file:
                packages = re.findall(r"[^\s]+", packages_file.read())

            package_system.install_packages(packages)

    def execute(self,
                argv,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                **kwargs):
        """Execute the process and arguments indicated by argv in container."""
        (argv,
         prepend_env,
         overwrite_env) = self._subprocess_popen_arguments(argv, **kwargs)

        with updated_environ(prepend_env, overwrite_env) as env:
            argv[0] = shutil.which(argv[0])
            assert argv[0] is not None

            # Also use which to find the shebang program - in some cases
            # we may only have the name of a program but not where it
            # actually exists. This is necessary on some platforms like
            # Windows where PATH is read from its state as it existed
            # when this process got created, not at the time Popen was
            # called.
            argv = parseshebang.parse(str(argv[0])) + argv
            argv[0] = shutil.which(argv[0])
            assert argv[0] is not None

            executed_cmd = subprocess.Popen(argv,
                                            stdout=stdout,
                                            stderr=stderr,
                                            env=env,
                                            universal_newlines=True)

        stdout_data, stderr_data = executed_cmd.communicate()

        return (executed_cmd.returncode, stdout_data, stderr_data)

    def execute_success(self, argv, **kwargs):
        """Execute the command specified by argv, throws on failure."""
        returncode, stdout_data, stderr_data = self.execute(argv,
                                                            subprocess.PIPE,
                                                            subprocess.PIPE,
                                                            **kwargs)

        if returncode != 0:
            sys.stderr.write(stdout_data)
            sys.stderr.write(stderr_data)
            raise RuntimeError("""{0} failed with {1}""".format(" ".join(argv),
                                                                returncode))



# /psqtraviscontainer/printer.py
#
# Utility functions for printing unicode text.
#
# See /LICENCE.md for Copyright information
"""Utility functions for printing unicode text."""

import platform

import sys


def unicode_safe(text):
    """Print text to standard output, handle unicode."""
    # If a replacement of sys.stdout doesn't have isatty, don't trust it.
    # Also don't trust Windows to get this right either.
    if (not getattr(sys.stdout, "isatty", None) or
            not sys.stdout.isatty() or
            platform.system() == "Windows"):
        text = "".join([c for c in str(text) if ord(c) < 128])

    sys.stdout.write(str(text))



# /psqtraviscontainer/osx_container.py
#
# Specialization for OS X containers, using environment variables.
#
# See /LICENCE.md for Copyright information
"""Specialization for OS X containers, using environment variables."""

import os

import platform

import shutil

import tarfile

from clint.textui import colored

from psqtraviscontainer import container
from psqtraviscontainer import directory
from psqtraviscontainer import distro
from psqtraviscontainer import package_system
from psqtraviscontainer import printer
from psqtraviscontainer import util

from psqtraviscontainer.download import TemporarilyDownloadedFile

import tempdir

DistroInfo = distro.DistroInfo

_HOMEBREW_URL = "https://github.com/Homebrew/homebrew/tarball/master"


class OSXContainer(container.AbstractContainer):
    """A container for OS X.

    We can execute commands inside this container by setting the
    required environment variables to pick commands from this
    path.
    """

    def __init__(self,  # suppress(too-many-arguments)
                 homebrew_distribution,
                 pkg_sys_constructor):
        """Initialize this OSXContainer, storing its distro configuration."""
        super(OSXContainer, self).__init__()
        self._prefix = homebrew_distribution
        self._pkgsys = pkg_sys_constructor(self)

    def _subprocess_popen_arguments(self, argv, **kwargs):
        """For native arguments argv, return AbstractContainer.PopenArguments.

        This returned tuple will have no environment variables set, but the
        proot command to enter this container will be prepended to the
        argv provided.
        """
        del kwargs

        popen_args = self.__class__.PopenArguments
        popen_env = {
            "PATH": os.path.join(self._prefix, "bin"),
            "LD_LIBRARY_PATH": os.path.join(self._prefix, "lib"),
            "PKG_CONFIG_PATH": os.path.join(self._prefix, "lib", "pkgconfig")
        }
        return popen_args(prepend=popen_env, argv=argv)

    def _root_filesystem_directory(self):
        """Return directory on parent filesystem where our root is located."""
        return self._prefix

    def _package_system(self):
        """Return package system for this distribution."""
        return self._pkgsys

    def clean(self):
        """Clean out this container to prepare it for caching."""
        pass


def _extract_archive(archive_file, container_folder):
    """Extract distribution archive into container_folder."""
    with tarfile.open(name=archive_file.path()) as archive:
        msg = ("""-> Extracting """
               """{0}\n""").format(archive_file.path())
        extract_members = archive.getmembers()
        printer.unicode_safe(colored.magenta(msg, bold=True))
        archive.extractall(members=extract_members, path=container_folder)


def container_for_directory(container_dir, distro_config):
    """Return an existing OSXContainer at container_dir for distro_config.

    Also take into account arguments in result to look up the the actual
    directory for this distro.
    """
    util.check_if_exists(os.path.join(container_dir, "bin", "brew"))

    return OSXContainer(container_dir, distro_config["pkgsys"])


def _fetch_homebrew(container_dir, distro_config):
    """Fetch homebrew and untar it in the container directory."""
    try:
        os.stat(os.path.join(container_dir, "bin", "brew"))
        return container_for_directory(container_dir, distro_config)
    except OSError:
        with directory.Navigation(tempdir.TempDir().name):
            with TemporarilyDownloadedFile(_HOMEBREW_URL) as archive_file:
                with directory.Navigation(tempdir.TempDir().name) as extract:
                    _extract_archive(archive_file, extract)
                    first = os.path.join(extract,
                                         os.listdir(extract)[0])
                    files = [os.path.join(first, p) for p in os.listdir(first)]
                    for filename in files:
                        try:
                            filename_base = os.path.basename(filename)
                            shutil.move(filename, os.path.join(container_dir,
                                                               filename_base))
                        except IOError:  # suppress(pointless-except)
                            # Ignore stuff that can't be moved for whatever
                            # reason. These are all files that generally
                            # don't matter.
                            pass

        return OSXContainer(container_dir, distro_config["pkgsys"])


def create(container_dir, distro_config):
    """Create a container using homebrew."""
    # First fetch a proot distribution if we don't already have one
    return _fetch_homebrew(container_dir, distro_config)


def match(info, arguments):
    """Check for matching configuration from DISTRIBUTIONS for arguments.

    In effect, this just means checking if we're on OS X.
    """
    if platform.system() != "Darwin":
        return None

    if arguments.get("distro", None) != "OSX":
        return None

    return info.kwargs


def enumerate_all(info):
    """Enumerate all valid configurations for this DistroInfo."""
    if platform.system() != "Darwin":
        return

    yield info.kwargs


DISTRIBUTIONS = [
    DistroInfo(create_func=create,
               get_func=container_for_directory,
               match_func=match,
               enumerate_func=enumerate_all,
               kwargs={
                   "distro": "OSX",
                   "pkgsys": package_system.Brew   # suppress(PYC50)
               })
]



# /psqtraviscontainer/windows_container.py
#
# Specialization for Windows containers, using environment variables.
#
# See /LICENCE.md for Copyright information
"""Specialization for Windows containers, using environment variables."""

import errno

import fnmatch

import os

import platform

import subprocess

import sys

from psqtraviscontainer import container
from psqtraviscontainer import directory
from psqtraviscontainer import distro
from psqtraviscontainer import package_system
from psqtraviscontainer import util

import tempdir

DistroInfo = distro.DistroInfo

_CHOCO_URL = "https://chocolatey.org/install.ps1"
_CHOCO_INSTALL_CMD = ("iex ((new-object net.webclient).DownloadString('" +
                      _CHOCO_URL + "'))")


class WindowsContainer(container.AbstractContainer):
    """A container for Windows.

    We can execute commands inside this container by setting the
    required environment variables to pick commands from this
    path.
    """

    def __init__(self,  # suppress(too-many-arguments)
                 chocolatey_distribution,
                 pkg_sys_constructor):
        """Initialize this WindowsContainer, storing its distro config."""
        super(WindowsContainer, self).__init__()
        self._prefix = chocolatey_distribution
        self._pkgsys = pkg_sys_constructor(self)

    def _subprocess_popen_arguments(self, argv):
        """For native arguments argv, return AbstractContainer.PopenArguments.

        This returned tuple will have no environment variables set, but the
        proot command to enter this container will be prepended to the
        argv provided.
        """
        popen_args = self.__class__.PopenArguments
        popen_env = {
            "PATH": os.path.join(self._prefix, "bin")
        }
        return popen_args(prepend=popen_env, argv=argv)

    def _root_filesystem_directory(self):
        """Return directory on parent filesystem where our root is located."""
        return self._prefix

    def _package_system(self):
        """Return package system for this distribution."""
        return self._pkgsys

    def clean(self):
        """Remove unnecessary files in this container."""
        rmtree = container.AbstractContainer.rmtree
        rmtree(os.path.join(self._prefix, "logs"))

        for root, directories, files in os.walk(os.path.join(self._prefix,
                                                             "lib")):
            for directory_name in directories:
                path_to_directory = os.path.join(root, directory_name)
                blacklist = [
                    "*/doc",
                    "*/man",
                    "*/html"
                ]

                for blacklisted_dir in blacklist:
                    if fnmatch.fnmatch(path_to_directory,
                                       blacklisted_dir):
                        rmtree(path_to_directory)

            for filename in files:
                path_to_file = os.path.join(root, filename)
                blacklist = [
                    "*.old"
                ]

                for blacklisted_file in blacklist:
                    if fnmatch.fnmatch(path_to_file,
                                       blacklisted_file):
                        os.remove(blacklisted_file)


def container_for_directory(container_dir, distro_config):
    """Return an existing WindowsContainer at container_dir for distro_config.

    Also take into account arguments in result to look up the the actual
    directory for this distro.
    """
    util.check_if_exists(os.path.join(container_dir, "bin", "choco.exe"))

    return WindowsContainer(container_dir, distro_config["pkgsys"])


def _execute_no_output(command):
    """Execute command, but don't show output unless it fails."""
    process = subprocess.Popen(command,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()

    if process.returncode != 0:
        sys.stdout.write(stdout)
        sys.stderr.write(stderr)
        raise RuntimeError("""Process {0} failed """
                           """with {1}""".format(" ".join(command),
                                                 process.returncode))


def _fetch_choco(container_dir, distro_config):
    """Fetch chocolatey and install it in the container directory."""
    try:
        os.stat(os.path.join(container_dir, "bin", "choco.exe"))
        return container_for_directory(container_dir, distro_config)
    except OSError:
        with tempdir.TempDir() as download_dir:
            with directory.Navigation(download_dir):
                _execute_no_output(["setx",
                                    "ChocolateyInstall",
                                    container_dir])

                # Also set the variable in the local environment
                # too, so that it gets propagated down to our
                # children
                os.environ["ChocolateyInstall"] = container_dir

                try:
                    os.makedirs(container_dir)
                except OSError as error:
                    if error.errno != errno.EEXIST:
                        raise error

                _execute_no_output(["powershell",
                                    "-NoProfile",
                                    "-ExecutionPolicy",
                                    "Bypass",
                                    "-Command",
                                    _CHOCO_INSTALL_CMD])

                # Reset variable back to original state to prevent
                # polluting the user's registry
                _execute_no_output(["setx",
                                    "ChocolateyInstall",
                                    ""])

        return WindowsContainer(container_dir, distro_config["pkgsys"])


def create(container_dir, distro_config):
    """Create a container using chocolatey."""
    return _fetch_choco(container_dir, distro_config)


def match(info, arguments):
    """Check for matching configuration from DISTRIBUTIONS for arguments.

    In effect, this just means checking if we're on Windows.
    """
    if platform.system() != "Windows":
        return None

    if arguments.get("distro", None) != "Windows":
        return None

    return info.kwargs


def enumerate_all(info):
    """Enumerate all valid configurations for this DistroInfo."""
    if platform.system() != "Windows":
        return

    yield info.kwargs


DISTRIBUTIONS = [
    DistroInfo(create_func=create,
               get_func=container_for_directory,
               match_func=match,
               enumerate_func=enumerate_all,
               kwargs={
                   "distro": "Windows",
                   "pkgsys": package_system.Choco
               })
]



# /psqtraviscontainer/linux_container.py
#
# Specialization for linux containers, using proot.
#
# See /LICENCE.md for Copyright information
"""Specialization for linux containers, using proot."""

from __future__ import unicode_literals

import errno

import os

import platform

import shutil

import stat

import tarfile

import tempfile

from collections import defaultdict
from collections import namedtuple

from contextlib import closing

from getpass import getuser

from itertools import chain

from clint.textui import colored

from psqtraviscontainer import architecture
from psqtraviscontainer import constants
from psqtraviscontainer import container
from psqtraviscontainer import directory
from psqtraviscontainer import distro
from psqtraviscontainer import package_system
from psqtraviscontainer import printer
from psqtraviscontainer import util

from psqtraviscontainer.download import TemporarilyDownloadedFile

import tempdir

_PROOT_URL_BASE = "http://static.proot.me/proot-{arch}"
_QEMU_URL_BASE = ("http://download.opensuse.org/repositories"
                  "/home:/cedric-vincent/xUbuntu_12.04/{arch}/"
                  "qemu-user-mode_1.6.1-1_{arch}.deb")


DistroInfo = distro.DistroInfo
DistroConfig = distro.DistroConfig
ProotDistribution = namedtuple("ProotDistribution", "proot qemu")


def proot_distro_from_container(container_dir):
    """Return a ProotDistribution from a container dir."""
    path_to_proot_dir = constants.proot_distribution_dir(container_dir)
    path_to_proot_bin = os.path.join(path_to_proot_dir, "bin/proot")
    path_to_qemu_template = os.path.join(path_to_proot_dir,
                                         "bin/qemu-{arch}")

    def _get_qemu_binary(arch):
        """Get the qemu binary for architecture."""
        qemu_arch = architecture.Alias.qemu(arch)
        return path_to_qemu_template.format(arch=qemu_arch)

    def _get_proot_binary():
        """Get the proot binary."""
        return path_to_proot_bin

    return ProotDistribution(proot=_get_proot_binary,
                             qemu=_get_qemu_binary)


def get_dir_for_distro(container_dir, config):
    """Get the distro dir in a container_dir for a DistroConfig."""
    arch = config["arch"]
    url = config["url"]
    distro_folder_name_template = (os.path.basename(url) + ".root")
    distro_folder_name = distro_folder_name_template.format(arch=arch)
    return os.path.realpath(os.path.join(container_dir, distro_folder_name))


def _rmtrees_as_container(cont, directories):
    """Remove directories as the root user in the container.

    This allows the removal of directories where permission errors
    might not permit otherwise.
    """
    root = cont.root_filesystem_directory()

    with tempfile.NamedTemporaryFile(dir=root, mode="wt") as bash_script:
        bash_script.write(";\n".join([("rm -rf " + d) for d in directories]))
        bash_script.flush()
        cont.execute(["bash", bash_script.name], minimal_bind=True)


class LinuxContainer(container.AbstractContainer):
    """A container for a linux distribution.

    We can execute commands inside this container by using proot and qemu.
    """

    def __init__(self,  # suppress(too-many-arguments)
                 proot_distribution,
                 distro_dir,
                 release,
                 arch,
                 pkg_sys_constructor):
        """Initialize this LinuxContainer, storing its distribution config."""
        super(LinuxContainer, self).__init__()
        self._proot_distro = proot_distribution
        self._distro_dir = distro_dir
        self._arch = arch
        self._pkgsys = pkg_sys_constructor(release, arch, self)

    def _subprocess_popen_arguments(self, argv, **kwargs):
        """For native arguments argv, return AbstractContainer.PopenArguments.

        This returned tuple will have no environment variables set, but the
        proot command to enter this container will be prepended to the
        argv provided.

        Pass minimal_bind=True to specify that no directories on the
        user filesystem should be exposed to the container. This will
        allow dpkg to remove certain system files in the container.
        """
        def parse_from_line(line):
            """Parse environment variable key-value pair from line."""
            return (line.split("=")[0],
                    "".join([c for c in line.split("=")[1]
                            if c != "\""]).strip())

        popen_args = self.__class__.PopenArguments

        if kwargs.get("minimal_bind", None):
            proot_command = [
                self._proot_distro.proot(),
                "-r",
                self._distro_dir,
                "-0"
            ]
        else:
            proot_command = [
                self._proot_distro.proot(),
                "-S",
                self._distro_dir
            ]

        # If we're not the same architecture, interpose qemu's emulator
        # for the target architecture as appropriate
        our_architecture = architecture.Alias.universal(platform.machine())
        target_architecture = architecture.Alias.universal(self._arch)

        if our_architecture != target_architecture:
            proot_command += ["-q", self._proot_distro.qemu(self._arch)]

        # Favor distribution's own environment variables
        with open(os.path.join(self._distro_dir, "etc", "environment")) as env:
            etc_environment_lines = env.readlines()
            prepend_env = dict([parse_from_line(l)
                                for l in etc_environment_lines
                                if l.split("=")[0].endswith("PATH")])
            overwrite_env = dict([parse_from_line(l)
                                 for l in etc_environment_lines
                                 if not l.split("=")[0].endswith("PATH")])

        # Make sure that LANG and LC_ALL are set to C, instead of
        # whatever it was set to before
        overwrite_env.update({
            "LANG": "C",
            "LC_ALL": "C"
        })

        return popen_args(prepend=prepend_env,
                          overwrite=overwrite_env,
                          argv=proot_command + argv)

    def _root_filesystem_directory(self):
        """Return directory on parent filesystem where our root is located."""
        return self._distro_dir

    def _package_system(self):
        """Return package system for this distribution."""
        return self._pkgsys

    def clean(self):
        """Clean out this container."""
        _rmtrees_as_container(self, [
            os.path.join(self._distro_dir, "tmp"),
            os.path.join(self._distro_dir, "var", "cache", "apt"),
            os.path.join(self._distro_dir, "var", "run"),
            os.path.join(self._distro_dir, "usr", "share", "doc"),
            os.path.join(self._distro_dir, "usr", "share", "locale"),
            os.path.join(self._distro_dir, "usr", "share", "man"),
            os.path.join(self._distro_dir, "var", "lib", "apt", "lists"),
            os.path.join(self._distro_dir, "dev")
        ])

        self.execute(["chown", "-R", "{}:users".format(getuser()), "/"],
                     minimal_bind=True)

        try:
            shutil.rmtree(os.path.join(self._distro_dir, "dev"))
        except OSError as error:
            if error.errno != errno.ENOENT:
                raise error

        try:
            os.makedirs(os.path.join(self._distro_dir,
                                     "var",
                                     "cache",
                                     "apt",
                                     "archives",
                                     "partial"))
        except OSError as error:
            if error.errno != errno.EEXIST:   # suppress(PYC90)
                raise error


def _extract_deb_data(archive, tmp_dir):
    """Extract archive to tmp_dir."""
    with closing(archive.getmember("data.tar.gz")) as member:
        with tarfile.open(fileobj=member,
                          mode="r|*") as data_tar:
            data_tar.extractall(path=tmp_dir)


def _fetch_proot_distribution(container_root):
    """Fetch the initial proot distribution if it is not available.

    Touches /.have-proot-distribution when complete
    """
    # We may not have python-debian installed on all platforms
    from debian import arfile  # suppress(import-error)

    path_to_proot_check = constants.have_proot_distribution(container_root)
    path_to_proot_dir = constants.proot_distribution_dir(container_root)

    def _download_proot(distribution_dir, arch):
        """Download arch build of proot into distribution."""
        from psqtraviscontainer.download import download_file

        with directory.Navigation(os.path.join(distribution_dir,
                                               "bin")):
            proot_url = _PROOT_URL_BASE.format(arch=arch)
            path_to_proot = download_file(proot_url, "proot")
            os.chmod(path_to_proot,
                     os.stat(path_to_proot).st_mode | stat.S_IXUSR)
            return path_to_proot

    def _extract_qemu(qemu_deb_path, qemu_temp_dir):
        """Extract qemu."""
        printer.unicode_safe(colored.magenta(("""-> Extracting {0}\n"""
                                              """""").format(qemu_deb_path),
                                             bold=True))
        archive = arfile.ArFile(qemu_deb_path)
        _extract_deb_data(archive, qemu_temp_dir)

    def _remove_unused_emulators(qemu_binaries_path):
        """Remove unused emulators from qemu distribution."""
        distributions = distro.available_distributions()
        cur_arch = platform.machine()
        archs = [d["info"].kwargs["arch"] for d in distributions]
        archs = set([architecture.Alias.qemu(a) for a in chain(*archs)
                     if a != architecture.Alias.universal(cur_arch)])
        keep_binaries = ["qemu-" + a for a in archs] + ["proot"]

        for root, _, filenames in os.walk(qemu_binaries_path):
            for filename in filenames:
                if os.path.basename(filename) not in keep_binaries:
                    os.remove(os.path.join(root, filename))

    def _download_qemu(distribution_dir, arch):
        """Download arch build of qemu and extract binaries."""
        qemu_url = _QEMU_URL_BASE.format(arch=arch)

        with TemporarilyDownloadedFile(qemu_url,
                                       filename="qemu.deb") as qemu_deb:
            # Go into a separate subdirectory and extract the qemu deb
            # there, then copy out the requisite files, so that we don't
            # cause tons of pollution
            qemu_tmp = os.path.join(path_to_proot_dir, "_qemu_tmp")
            with directory.Navigation(qemu_tmp):
                qemu_binaries_path = os.path.join(qemu_tmp, "usr", "bin")
                _extract_qemu(qemu_deb.path(), qemu_tmp)
                _remove_unused_emulators(qemu_binaries_path)

                for filename in os.listdir(qemu_binaries_path):
                    shutil.copy(os.path.join(qemu_binaries_path, filename),
                                os.path.join(path_to_proot_dir, "bin"))

            shutil.rmtree(qemu_tmp)

        return os.path.join(distribution_dir, "bin", "qemu-{arch}")

    try:
        os.stat(path_to_proot_check)
        printer.unicode_safe(colored.green("""-> """
                                           """Using pre-existing proot """
                                           """distribution\n""",
                                           bold=True))

    except OSError:
        create_msg = """Creating distribution of proot in {}\n"""
        root_relative = os.path.relpath(container_root)
        printer.unicode_safe(colored.yellow(create_msg.format(root_relative),
                                            bold=True))

        # Distro check does not exist - create the ./_proot directory
        # and download files for this architecture
        with directory.Navigation(path_to_proot_dir):
            proot_arch = architecture.Alias.universal(platform.machine())
            qemu_arch = architecture.Alias.debian(platform.machine())
            _download_proot(path_to_proot_dir, proot_arch)
            _download_qemu(path_to_proot_dir, qemu_arch)

        with open(path_to_proot_check, "w+") as check_file:
            check_file.write("done")

        printer.unicode_safe(colored.green("""\N{check mark} """
                                           """Successfully installed proot """
                                           """distribution to """
                                           """{}\n""".format(root_relative),
                                           bold=True))

    return proot_distro_from_container(container_root)


def _extract_distro_archive(distro_archive_file, distro_folder):
    """Extract distribution archive into distro_folder."""
    with tarfile.open(name=distro_archive_file.path()) as archive:
        msg = ("""-> Extracting """
               """{0}\n""").format(os.path.relpath(distro_archive_file.path()))
        extract_members = [m for m in archive.getmembers()
                           if not m.isdev()]
        printer.unicode_safe(colored.magenta(msg, bold=True))
        archive.extractall(members=extract_members, path=distro_folder)

        # Set the permissions of the extracted archive so we can delete it
        # if need be.
        os.chmod(distro_folder, os.stat(distro_folder).st_mode | stat.S_IRWXU)
        for root, directories, filenames in os.walk(distro_folder):
            for distro_folder_directory in directories:
                path = os.path.join(root, distro_folder_directory)
                try:
                    os.chmod(path, os.stat(path).st_mode | stat.S_IRWXU)
                except OSError:  # suppress(pointless-except)
                    pass
            for filename in filenames:
                path = os.path.join(root, filename)
                try:
                    os.chmod(path, os.stat(path).st_mode | stat.S_IRWXU)
                except OSError:   # suppress(pointless-except)
                    pass


def _fetch_distribution(container_root,  # pylint:disable=R0913
                        proot_distro,
                        details):
    """Lazy-initialize distribution and return it."""
    path_to_distro_folder = get_dir_for_distro(container_root,
                                               details)

    def _download_distro(details, path_to_distro_folder):
        """Download distribution and untar it in container root."""
        distro_arch = details["arch"]
        download_url = details["url"].format(arch=distro_arch)
        with tempdir.TempDir() as download_dir:
            with directory.Navigation(download_dir):
                with TemporarilyDownloadedFile(download_url) as archive_file:
                    _extract_distro_archive(archive_file,
                                            path_to_distro_folder)

    def _minimize_ubuntu(cont):
        """Reduce the install footprint of ubuntu as much as possible."""
        required_packages = {
            "precise": set([
                "apt",
                "base-files",
                "base-passwd",
                "bash",
                "bsdutils",
                "coreutils",
                "dash",
                "debconf",
                "debianutils",
                "diffutils",
                "dpkg",
                "findutils",
                "gcc-4.6-base",
                "gnupg",
                "gpgv",
                "grep",
                "gzip",
                "libacl1",
                "libapt-pkg4.12",
                "libattr1",
                "libbz2-1.0",
                "libc-bin",
                "libc6",
                "libdb5.1",
                "libffi6",
                "libgcc1",
                "liblzma5",
                "libpam-modules",
                "libpam-modules-bin",
                "libpam-runtime",
                "libpam0g",
                "libreadline6",
                "libselinux1",
                "libstdc++6",
                "libtinfo5",
                "libusb-0.1-4",
                "makedev",
                "mawk",
                "multiarch-support",
                "perl-base",
                "readline-common",
                "sed",
                "sensible-utils",
                "tar",
                "tzdata",
                "ubuntu-keyring",
                "xz-utils",
                "zlib1g"
            ]),
            "trusty": set([
                "apt",
                "base-files",
                "base-passwd",
                "bash",
                "bsdutils",
                "coreutils",
                "dash",
                "debconf",
                "debianutils",
                "diffutils",
                "dh-python",
                "dpkg",
                "findutils",
                "gcc-4.8-base",
                "gcc-4.9-base",
                "gnupg",
                "gpgv",
                "grep",
                "gzip",
                "libacl1",
                "libapt-pkg4.12",
                "libaudit1",
                "libaudit-common",
                "libattr1",
                "libbz2-1.0",
                "libc-bin",
                "libc6",
                "libcap2",
                "libdb5.3",
                "libdebconfclient0",
                "libexpat1",
                "libmpdec2",
                "libffi6",
                "libgcc1",
                "liblzma5",
                "libncursesw5",
                "libpcre3",
                "libpam-modules",
                "libpam-modules-bin",
                "libpam-runtime",
                "libpam0g",
                "libpython3-stdlib",
                "libpython3.4-stdlib",
                "libpython3",
                "libpython3-minimal",
                "libpython3.4",
                "libpython3.4-minimal",
                "libreadline6",
                "libselinux1",
                "libssl1.0.0",
                "libstdc++6",
                "libsqlite3-0",
                "libtinfo5",
                "libusb-0.1-4",
                "lsb-release",
                "makedev",
                "mawk",
                "mime-support",
                "multiarch-support",
                "perl-base",
                "python3",
                "python3-minimal",
                "python3.4",
                "python3.4-minimal",
                "readline-common",
                "sed",
                "sensible-utils",
                "tar",
                "tzdata",
                "ubuntu-keyring",
                "xz-utils",
                "zlib1g"
            ])
        }

        os.environ["SUDO_FORCE_REMOVE"] = "yes"
        os.environ["DEBIAN_FRONTEND"] = "noninteractive"

        pkgs = set(cont.execute(["dpkg-query",
                                 "-Wf",
                                 "${Package} "])[1].split(" "))
        release = details["release"]
        remove = [l for l in list(pkgs ^ required_packages[release]) if len(l)]

        if len(remove):
            cont.execute_success(["dpkg",
                                  "--purge",
                                  "--force-all"] + remove,
                                 minimal_bind=True)

        with open(os.path.join(get_dir_for_distro(container_root,
                                                  details),
                               "etc",
                               "apt",
                               "apt.conf.d",
                               "99container"), "w") as apt_config:
            apt_config.write("\n".join([
                "APT::Install-Recommends \"0\";",
                "APT::Install-Suggests \"0\";"
            ]))

    # Container isn't safe to use until we've either verified that the
    # path to the distro folder exists or we've downloaded a distro into it
    linux_cont = LinuxContainer(proot_distro,
                                path_to_distro_folder,
                                details["release"],
                                details["arch"],
                                details["pkgsys"])

    try:
        os.stat(path_to_distro_folder)
        use_existing_msg = ("""\N{check mark} Using existing folder for """
                            """proot distro """
                            """{distro} {release} {arch}\n""")
        printer.unicode_safe(colored.green(use_existing_msg.format(**details),
                                           bold=True))
    except OSError:
        # Download the distribution tarball in the distro dir
        _download_distro(details, path_to_distro_folder)

        # Minimize the installed distribution, but only when it
        # was just initially downloaded
        minimize_actions = defaultdict(lambda: lambda c: None,
                                       Ubuntu=_minimize_ubuntu)
        minimize_actions[details["distro"]](linux_cont)

    return linux_cont


def container_for_directory(container_dir, distro_config):
    """Return an existing LinuxContainer at container_dir for distro_config.

    Also take into account arguments in result to look up the the actual
    directory for this distro.
    """
    path_to_distro_folder = get_dir_for_distro(container_dir,
                                               distro_config)

    required_entities = [
        constants.have_proot_distribution(container_dir),
        path_to_distro_folder
    ]

    for entity in required_entities:
        util.check_if_exists(entity)

    proot_distribution = proot_distro_from_container(container_dir)

    return LinuxContainer(proot_distribution,
                          path_to_distro_folder,
                          distro_config["release"],
                          distro_config["arch"],
                          distro_config["pkgsys"])


def create(container_dir, distro_config):
    """Create a container using proot."""
    # First fetch a proot distribution if we don't already have one
    proot_distro = _fetch_proot_distribution(container_dir)

    # Now fetch the distribution tarball itself, if we specified one
    cont = _fetch_distribution(container_dir,
                               proot_distro,
                               distro_config)

    return cont


def _info_with_arch_to_config(info, arch):
    """Convert selected architecture for DistroInfo into DistroConfig."""
    config = info.kwargs.copy()

    del config["arch"]
    del config["archfetch"]

    config["arch"] = arch

    return config


def _valid_archs(archs):
    """Return valid archs to emulate from archs.

    64 bit architectures can't be emulated on a 32 bit system, so remove
    them form the list of valid architectures.
    """
    blacklist = defaultdict(lambda: None)
    blacklist["x86"] = "x86_64"
    blacklist["x86_64"] = "x86"

    arch_alias = architecture.Alias.universal
    machine = arch_alias(platform.machine())

    return [a for a in archs if arch_alias(a) != blacklist[machine]]


def match(info, arguments):
    """Check if info matches arguments."""
    if platform.system() != "Linux":
        return None

    if arguments.get("distro", None) != info.kwargs["distro"]:
        return None

    distro_release = info.kwargs["release"]

    # pychecker thinks that a list comprehension as a return value is
    # always None.
    distro_archs = _valid_archs(info.kwargs["arch"])  # suppress(PYC90)
    distro_archfetch = info.kwargs["archfetch"]

    if arguments.get("release", None) == distro_release:
        converted = distro_archfetch(arguments.get("arch", None))
        if converted in distro_archs:
            return _info_with_arch_to_config(info, converted)

    return None


def enumerate_all(info):
    """Enumerate all valid configurations for this DistroInfo."""
    if platform.system() != "Linux":
        return

    for arch in _valid_archs(info.kwargs["arch"]):  # suppress(PYC90)
        yield _info_with_arch_to_config(info, arch)


class LinuxInfo(DistroInfo):
    """Linux-specific specialization of DistroInfo."""

    PACKAGE_SYSTEMS = {
        "Ubuntu": package_system.Dpkg,
        "Debian": package_system.Dpkg,
        "Fedora": package_system.Yum
    }

    def __new__(cls, distro_type, **kwargs):
        """Create DistroInfo namedtuple using provided arguments."""
        kwargs.update({
            "distro": distro_type,
            "pkgsys": LinuxInfo.PACKAGE_SYSTEMS[distro_type]
        })

        return DistroInfo.__new__(cls,
                                  create_func=create,
                                  get_func=container_for_directory,
                                  match_func=match,
                                  enumerate_func=enumerate_all,
                                  kwargs=kwargs)

DISTRIBUTIONS = [  # suppress(unused-variable)
    LinuxInfo("Ubuntu",
              release="precise",
              url=("http://cdimage.ubuntu.com/ubuntu-core/releases/"
                   "precise/release/ubuntu-core-12.04.5-core-{arch}.tar.gz"),
              arch=["i386", "amd64", "armhf"],
              archfetch=architecture.Alias.debian),
    LinuxInfo("Ubuntu",
              release="trusty",
              url=("http://cdimage.ubuntu.com/ubuntu-core/releases/"
                   "trusty/release/ubuntu-core-14.04.1-core-{arch}.tar.gz"),
              arch=["i386", "amd64", "armhf", "powerpc"],
              archfetch=architecture.Alias.debian),
    LinuxInfo("Debian",
              release="wheezy",
              url=("http://download.openvz.org/"
                   "template/precreated/debian-7.0-{arch}-minimal.tar.gz"),
              arch=["x86", "x86_64"],
              archfetch=architecture.Alias.universal),
    LinuxInfo("Debian",
              release="squeeze",
              url=("http://download.openvz.org/"
                   "template/precreated/debian-6.0-{arch}-minimal.tar.gz"),
              arch=["x86", "x86_64"],
              archfetch=architecture.Alias.universal),
    LinuxInfo("Fedora",
              release="20",
              url=("http://download.openvz.org/"
                   "template/precreated/fedora-20-{arch}.tar.gz"),
              arch=["x86", "x86_64"],
              # suppress(PYC50)
              archfetch=architecture.Alias.universal)
]



# /psqtraviscontainer/constants.py
#
# Constants shared across modules.
#
# See /LICENCE.md for Copyright information
"""Various constants useful for use and create modules."""

import os


def have_proot_distribution(cwd):
    """Return proot distribution stamp filename."""
    return os.path.join(cwd, ".have-proot-distribution")


def proot_distribution_dir(cwd):
    """Return proot distribution dir from cwd."""
    return os.path.join(cwd, "_proot")



# /psqtraviscontainer/package_system.py
#
# Implementations of package-system controllers for various distributions.
#
# See /LICENCE.md for Copyright information
"""Implementations of package-system controllers for various distributions."""

import abc

import sys

import tempfile

from collections import namedtuple

from clint.textui import colored

from psqtraviscontainer import directory
from psqtraviscontainer import download

import six

import tempdir

_UBUNTU_MAIN_ARCHS = ["i386", "amd64"]
_UBUNTU_PORT_ARCHS = ["armhf", "arm64", "powerpc", "ppc64el"]
_UBUNTU_MAIN_ARCHIVE = "http://archive.ubuntu.com/ubuntu/"
_UBUNTU_PORT_ARCHIVE = "http://ports.ubuntu.com/ubuntu-ports/"


def _run_task(executor, description, argv):
    """Run command through executor argv and prints description."""
    sys.stdout.write(str(colored.white("-> {0}\n".format(description))))
    executor.execute_success(argv)


class PackageSystem(six.with_metaclass(abc.ABCMeta, object)):
    """An abstract class representing a package manager."""

    PopenArguments = namedtuple("PopenArguments", "argv env")

    @abc.abstractmethod
    def add_repositories(self, repos):
        """Add repositories to central packaging system."""
        del repos

        raise NotImplementedError()

    @abc.abstractmethod
    def install_packages(self, package_names):
        """Install specified packages in package_names."""
        del package_names

        raise NotImplementedError()


class Dpkg(PackageSystem):
    """Debian Packaging System."""

    def __init__(self,
                 release,
                 arch,
                 executor):
        """Initialize Dpkg with release and arch."""
        super(Dpkg, self).__init__()
        self._release = release
        self._arch = arch
        self._executor = executor

    def add_repositories(self, repos):
        """Add a repository to the central packaging system."""
        _ubuntu_urls = [
            (_UBUNTU_MAIN_ARCHS, _UBUNTU_MAIN_ARCHIVE),
            (_UBUNTU_PORT_ARCHS, _UBUNTU_PORT_ARCHIVE)
        ]

        def _format_user_line(line, kwargs):
            """Format a line and turns it into a valid repo line."""
            formatted_line = line.format(**kwargs)
            return "deb {0}".format(formatted_line)

        def _value_or_error(value):
            """Return first item in value, or ERROR if value is empty."""
            return value[0] if len(value) else "ERROR"

        format_keys = {
            "ubuntu": [u[1] for u in _ubuntu_urls if self._arch in u[0]],
            "debian": ["http://ftp.debian.org/"],
            "launchpad": ["http://ppa.launchpad.net/"],
            "release": [self._release]
        }
        format_keys = {
            k: _value_or_error(v) for k, v in format_keys.items()
        }

        # We will be creating a bash script each time we need to add
        # a new source line to our sources list and executing that inside
        # the proot. This guarantees that we'll always get the right
        # permissions.
        with tempfile.NamedTemporaryFile() as bash_script:
            append_lines = [_format_user_line(l, format_keys) for l in repos]
            for count, append_line in enumerate(append_lines):
                path = "/etc/apt/sources.list.d/{0}.list".format(count)
                append_cmd = "echo \"{0}\" > {1}\n".format(append_line, path)
                bash_script.write(six.b(append_cmd))

            bash_script.flush()
            self._executor.execute_success(["bash", bash_script.name])

    def install_packages(self, package_names):
        """Install all packages in list package_names."""
        _run_task(self._executor,
                  """Update repositories""",
                  ["apt-get", "update", "-qq", "-y", "--force-yes"])
        _run_task(self._executor,
                  """Install {0}""".format(str(package_names)),
                  ["apt-get",
                   "install",
                   "-qq",
                   "-y",
                   "--force-yes"] + package_names)


class Yum(PackageSystem):
    """Red Hat Packaging System."""

    def __init__(self,
                 release,
                 arch,
                 executor):
        """Initialize Yum with release and executor."""
        del arch
        del release

        super(Yum, self).__init__()
        self._executor = executor

    def add_repositories(self, repos):
        """Add a repository to the central packaging system."""
        with tempdir.TempDir() as download_dir:
            with directory.Navigation(download_dir):
                for repo in repos:
                    repo_file = download.download_file(repo)
                    # Create a bash script to copy the downloaded repo file
                    # over to /etc/yum/repos.d
                    with tempfile.NamedTemporaryFile() as bash_script:
                        copy_cmd = ("cp \"{0}\""
                                    "/etc/yum/repos.d").format(repo_file)
                        bash_script.write(six.b(copy_cmd))
                        bash_script.flush()
                        self._executor.execute_success(["bash",
                                                        bash_script.name])

    def install_packages(self, package_names):
        """Install all packages in list package_names."""
        _run_task(self._executor,
                  """Install {0}""".format(str(package_names)),
                  ["yum", "install", "-y"] + package_names)


class Brew(PackageSystem):
    """Homebrew packaging system for OS X."""

    def __init__(self, executor):
        """Initialize homebrew for executor."""
        super(Brew, self).__init__()
        self._executor = executor

    def add_repositories(self, repos):
        """Add repositories as specified at repos.

        Adds repositories using brew tap.
        """
        for repo in repos:
            _run_task(self._executor,
                      """Adding repository {0}""".format(repo),
                      ["brew", "tap", repo])

    def install_packages(self, package_names):
        """Install all packages in list package_names."""
        _run_task(self._executor,
                  """Install {0}""".format(str(package_names)),
                  ["brew", "install"] + package_names)


class Choco(PackageSystem):
    """Chocolatey packaging system for Windows."""

    def __init__(self, executor):
        """Initialize choco for executor."""
        super(Choco, self).__init__()
        self._executor = executor

    def add_repositories(self, repos):
        """Add repositories as specified at repos.

        This function doesn't do anything on Choco at the moment.
        """
        pass

    def install_packages(self, package_names):
        """Install all packages in list package_names."""
        _run_task(self._executor,
                  """Install {0}""".format(str(package_names)),
                  ["choco", "install", "-fy", "-m"] + package_names)



# /test/test_unit.py
#
# Unit tests for various utilities.
#
# See /LICENCE.md for Copyright information
"""Unit tests for various utilities."""

import os

import shutil

from psqtraviscontainer import architecture
from psqtraviscontainer import directory
from psqtraviscontainer import distro

from testtools import ExpectedException
from testtools import TestCase

from testtools.matchers import AllMatch
from testtools.matchers import DirExists
from testtools.matchers import MatchesPredicate


class TestDirectoryNavigation(TestCase):
    """Tests for psqtraviscontainer/directory.py."""

    def test_enter_create_dir(self):
        """Check that we create a dir when entering a non-existent one."""
        does_not_exist = os.path.join(os.getcwd(), "does_not_exist")
        self.addCleanup(lambda: shutil.rmtree(does_not_exist))
        with directory.Navigation(does_not_exist) as entered:
            self.assertThat(entered, DirExists())

    def test_enter_exist_dir(self):
        """Check that we can enter an existing dir."""
        existing_dir = os.path.join(os.getcwd(), "existing")
        os.makedirs(existing_dir)
        self.addCleanup(lambda: shutil.rmtree(existing_dir))
        with directory.Navigation(existing_dir) as entered:
            self.assertThat(entered, DirExists())


class TestArchitecture(TestCase):  # suppress(R0903)
    """Tests for psqtraviscontainer/architecture.py."""

    def test_unknown_architecture(self):
        """Check that creating a non-special architecture returns metadata."""
        check_methods = [
            architecture.Alias.universal,
            architecture.Alias.qemu,
            architecture.Alias.debian
        ]

        def function_returns_input(function):
            """Return true if function returns input."""
            return function("input") == "input"

        self.assertThat(check_methods,
                        AllMatch(MatchesPredicate(function_returns_input,
                                                  "% did not return same")))


class TestDistroLookup(TestCase):  # suppress(R0903)
    """Tests for looking up the distro."""

    def test_error_lookup_bad_distro(self):  # suppress(no-self-use)
        """Check that looking up a non-existent distro throws."""
        with ExpectedException(RuntimeError):
            distro.lookup({"distro": "noexist"})



# /psqtraviscontainer/rootdir.py
#
# Print the root directory of a selected container to standard out.
#
# See /LICENCE.md for Copyright information
"""Print the root directory of a selected container to standard out."""

import os

import sys

from psqtraviscontainer import common_options
from psqtraviscontainer import distro


def main(arguments=None):
    """Get container and print root filesystem directory."""
    parser = common_options.get_parser("""Get root directory for""")
    result = parser.parse_args(arguments)
    container_dir = os.path.realpath(result.containerdir)

    selected_distro = distro.lookup(vars(result))

    # Get the selected distribution's container and print its root
    # filesystem directory
    with selected_distro["info"].get_func(container_dir,
                                          selected_distro) as container:
        sys.stdout.write(container.root_filesystem_directory())

if __name__ == "__main__":
    main()



# /test/test_acceptance.py
#
# Test case for psqtraviscontainer/create.py, creating proot containers
#
# See /LICENCE.md for Copyright information
"""Test case for psqtraviscontainer/create.py, creating proot containers."""

import os

import platform

import stat

import sys

import tempfile

from collections import namedtuple

from contextlib import contextmanager

from test.testutil import download_file_cached

from nose_parameterized import parameterized

from psqtraviscontainer import architecture
from psqtraviscontainer import create
from psqtraviscontainer import use

from psqtraviscontainer.architecture import Alias

from psqtraviscontainer.constants import have_proot_distribution
from psqtraviscontainer.constants import proot_distribution_dir

from psqtraviscontainer.distro import available_distributions

from psqtraviscontainer.linux_container import get_dir_for_distro

import tempdir

from testtools import ExpectedException
from testtools import TestCase

from testtools.matchers import DirExists
from testtools.matchers import FileExists


def _convert_to_switch_args(kwargs):
    """Convert keyword arguments to command line switches."""
    arguments = []

    def _get_representation(value):
        """Get representation of value as a list."""
        if isinstance(value, list):
            return " ".join(value)
        else:
            return str(value)

    for key, value in kwargs.items():
        arguments.append("--{0}".format(key))
        arguments.append(_get_representation(value))

    return arguments


class SafeTempDir(object):  # pylint:disable=R0903
    """A TempDir that dissolves on __exit__, ignoring PermissionError."""

    def __init__(self):
        """Forward initialization."""
        super(SafeTempDir, self).__init__()
        self._temp_dir = tempdir.TempDir()

    def __enter__(self):
        """Return internal tempdir."""
        return self._temp_dir.__enter__()

    def __exit__(self, exc_type, value, traceback):
        """Call dissolve."""
        del exc_type
        del value
        del traceback

        self.dissolve()

    def dissolve(self):
        """Forward to TempDir dissolve function, ignore PermissionError."""
        try:
            self._temp_dir.dissolve()
        except (IOError, OSError):  # suppress(pointless-except)
            # IOError and OSError are fine. The directory will be deleted by
            # the user's operating system a little later, there's not much we
            # can do about this.
            pass

    @property
    def name(self):
        """Getter for 'name'."""
        return self._temp_dir.name


def run_create_container_on_dir(directory, *args, **kwargs):
    """Run main setting the container to be at directory."""
    del args

    arguments = [directory] + _convert_to_switch_args(kwargs)

    with cached_downloads():
        create.main(arguments=arguments)


def run_create_container(**kwargs):
    """Run main() and returns the container in a TempDir.

    This houses the container created. Keyword args are converted
    into switch arguments as appropriate.
    """
    temp_dir = SafeTempDir()
    run_create_container_on_dir(temp_dir.name, **kwargs)
    return temp_dir


def default_create_container_arguments():
    """Get set of arguments which would create first known distribution."""
    distro_config = list(available_distributions())[0]
    arguments = ("distro", "release")
    config = {k: v for k, v in distro_config.items() if k in arguments}
    return config


def run_create_default_container():
    """Run main() and return container for first known distribution."""
    return run_create_container(**(default_create_container_arguments()))


def run_use_container_on_dir(directory, **kwargs):
    """Run main() from psqtraviscontainer/use.py and return status code."""
    cmd = kwargs["cmd"]
    del kwargs["cmd"]

    arguments = [directory] + _convert_to_switch_args(kwargs) + ["--"] + cmd

    return use.main(arguments=arguments)


def test_case_requiring_platform(system):
    """Get a TestCase base class which can only be run on platform."""
    class TestCaseRequiring(TestCase):
        """A wrapper around TestCase which only runs tests on platform."""

        def setUp(self):  # suppress(N802)
            """Automatically skips tests if not run on platform."""
            super(TestCaseRequiring, self).setUp()
            if platform.system() != system:
                self.skipTest("""not running on system - {0}""".format(system))

    return TestCaseRequiring


class TestCreateProot(test_case_requiring_platform("Linux")):
    """A test case for proot creation basics."""

    def test_create_proot_distro(self):
        """Check that we create a proot distro."""
        with run_create_default_container() as container:
            self.assertThat(have_proot_distribution(container),
                            FileExists())

    def test_use_existing_proot_distro(self):
        """Check that we re-use an existing proot distro.

        In that case, the timestamp for /.have-proot-distribution and
        make sure that across two runs they are actual. If they were,
        then no re-downloading took place.
        """
        with run_create_default_container() as container:
            path_to_proot_stamp = have_proot_distribution(container)

            first_timestamp = os.stat(path_to_proot_stamp).st_mtime

            config = default_create_container_arguments()
            run_create_container_on_dir(container, **config)

            second_timestamp = os.stat(path_to_proot_stamp).st_mtime

            self.assertEqual(first_timestamp, second_timestamp)


@contextmanager
def cached_downloads():
    """Context manager to ensure that download_file is patched to use cache."""
    import six
    import psqtraviscontainer.download  # suppress(PYC50)

    original_download_file = psqtraviscontainer.download.download_file
    psqtraviscontainer.download.download_file = download_file_cached

    original_stdout = sys.stdout
    original_stderr = sys.stderr

    if not os.environ.get("_POLYSQUARE_TRAVIS_CONTAINER_TEST_SHOW_OUTPUT",
                          None):
        sys.stdout = six.StringIO()
        sys.stderr = six.StringIO()

    try:
        yield
    finally:
        psqtraviscontainer.download.download_file = original_download_file
        sys.stdout = original_stdout
        sys.stderr = original_stderr


class ContainerInspectionTestCase(TestCase):
    """TestCase where container persists until all tests have completed.

    No modifications should be made to the container during any
    individual test. The order of tests should not be relied upon.
    """

    container_temp_dir = None

    def __init__(self, *args, **kwargs):
        """Initialize class."""
        cls = ContainerInspectionTestCase
        super(cls, self).__init__(*args, **kwargs)
        self.container_dir = None

    def setUp(self):  # suppress(N802)
        """Set up container dir."""
        super(ContainerInspectionTestCase, self).setUp()
        self.container_dir = self.__class__.container_temp_dir.name

    @classmethod
    def create_container(cls, **kwargs):
        """Overridable method to create a container for this test case."""
        cls.container_temp_dir = run_create_container(**kwargs)

    # Suppress flake8 complaints about uppercase characters in function names,
    # these functions are overloaded
    @classmethod
    def setUpClass(cls):  # suppress(N802)
        """Set up container for all tests in this test case."""
        config = default_create_container_arguments()
        cls.create_container(**config)

    @classmethod
    def tearDownClass(cls):  # suppress(N802)
        """Dissolve container for all tests in this test case."""
        if cls.container_temp_dir:
            cls.container_temp_dir.dissolve()
            cls.container_temp_dir = None


QEMU_ARCHITECTURES = [
    "arm",
    "i386",
    "ppc",
    "x86_64"
]


def _format_arch(func, num, params):
    """Format docstring for TestProotDistribution parameterized tests."""
    del num

    return func.__doc__.format(arch=params[0][0])


class TestProotDistribution(ContainerInspectionTestCase):
    """Tests to inspect a proot distribution itself."""

    def setUp(self):   # suppress(N802)
        """Set up TestProotDistribution."""
        if platform.system() != "Linux":
            self.skipTest("""proot is only available on linux""")

        super(TestProotDistribution, self).setUp()

    def test_has_proot_dir(self):
        """Check that we have a proot directory in our distribution."""
        self.assertThat(proot_distribution_dir(self.container_dir),
                        DirExists())

    def test_has_proot_executable(self):
        """Check that we have a proot executable in our distribution."""
        cont = proot_distribution_dir(self.container_dir)
        self.assertThat(os.path.join(cont, "bin/proot"),
                        FileExists())

    def test_proot_binary_is_executable(self):
        """Check that that the proot binary is executable."""
        cont = proot_distribution_dir(self.container_dir)
        proot_binary = os.path.join(cont, "bin/proot")
        stat_result = os.stat(proot_binary)
        executable_mask = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
        self.assertTrue(stat_result.st_mode & executable_mask != 0)

    @parameterized.expand(QEMU_ARCHITECTURES, testcase_func_doc=_format_arch)
    def test_has_qemu_executables(self, arch):
        """Check that we have a qemu executable qemu-{arch}."""
        cont = proot_distribution_dir(self.container_dir)
        self.assertThat(os.path.join(cont, "bin/qemu-{}".format(arch)),
                        FileExists())

    @parameterized.expand(QEMU_ARCHITECTURES, testcase_func_doc=_format_arch)
    def test_qemu_binary_is_executable(self, arch):
        """Check that qemu binary qemu-{arch} is executable."""
        cont = proot_distribution_dir(self.container_dir)
        proot_binary = os.path.join(cont, "bin/qemu-{}".format(arch))
        stat_result = os.stat(proot_binary)
        executable_mask = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
        self.assertTrue(stat_result.st_mode & executable_mask != 0)


def exec_for_returncode(*argv):
    """Execute command for its return code.

    Check that use.main() returns exit code of subprocess.
    """
    config = default_create_container_arguments()
    with run_create_container(**config) as cont:
        use_config = config.copy()
        use_config["cmd"] = list(argv)
        return run_use_container_on_dir(cont,
                                        **use_config)


PLATFORM_PROGRAM_MAPPINGS = {
    "Linux": {
        "0": ["true"],
        "1": ["false"]
    },
    "Darwin": {
        "0": ["true"],
        "1": ["false"]
    },
    "Windows": {
        "0": ["python", "-c", "import sys;sys.exit(0);"],
        "1": ["python", "-c", "import sys;sys.exit(1);"]
    }
}


class TestExecInContainer(TestCase):
    """A test case for executing things inside a container."""

    def test_exec_fail_no_distro(self):  # suppress(no-self-use)
        """Check that use.main() fails where there is no distro."""
        with run_create_default_container() as container_dir:
            with ExpectedException(RuntimeError):
                cmd = PLATFORM_PROGRAM_MAPPINGS[platform.system()]["0"]
                run_use_container_on_dir(container_dir, cmd=cmd)

    def test_exec_return_zero(self):
        """Check that use.main() returns true exit code of subprocess."""
        cmd = PLATFORM_PROGRAM_MAPPINGS[platform.system()]["0"]
        self.assertEqual(exec_for_returncode(*cmd), 0)

    def test_exec_return_one(self):
        """Check that use.main() returns false exit code of subprocess."""
        cmd = PLATFORM_PROGRAM_MAPPINGS[platform.system()]["1"]
        self.assertEqual(exec_for_returncode(*cmd), 1)

ARCHITECTURE_LIBDIR_MAPPINGS = {
    "armhf": "arm-linux-gnueabihf",
    "i386": "i386-linux-gnu",
    "amd64": "x86_64-linux-gnu",
    "arm64": "arm64-linux-gnu",
    "powerpc": "powerpc-linux-gnu",
    "ppc64el": "ppc64el-linux-gnu"
}


class InstallationConfig(object):  # pylint:disable=R0903
    """Manages configuration files."""

    def __init__(self, packages, repos):
        """Create temporary files for packages and repos."""
        packages_fd, self.packages_path = tempfile.mkstemp()
        repos_fd, self.repos_path = tempfile.mkstemp()

        packages_file = os.fdopen(packages_fd, "a")
        repos_file = os.fdopen(repos_fd, "a")

        for package in packages:
            packages_file.write("{0}\n".format(package))

        for repository in repos:
            repos_file.write("{0}\n".format(repository))

        packages_file.close()
        repos_file.close()

    def __enter__(self):
        """Use as context manager."""
        return self

    def __exit__(self, exc_type, value, traceback):
        """Destroy temporary files."""
        del exc_type
        del value
        del traceback

        os.remove(self.packages_path)
        os.remove(self.repos_path)


def _create_distro_test(test_name,  # pylint:disable=R0913
                        config,
                        repos,
                        packages,
                        test_files,
                        **kwargs):
    """Create a TemplateDistroTest class."""
    class TemplateDistroTest(ContainerInspectionTestCase):
        """Template for checking a distro proot."""

        def __init__(self, *args, **kwargs):
            """Initialize members used by this class."""
            cls = TemplateDistroTest
            super(cls, self).__init__(*args, **kwargs)
            self.path_to_distro_root = None

        def setUp(self):  # suppress(N802)
            """Set up path to distro root."""
            super(TemplateDistroTest, self).setUp()

        @classmethod
        def setUpClass(cls):  # suppress(N802)
            """Create a container for all uses of this TemplateDistroTest."""
            with InstallationConfig(packages, repos) as command_config:
                keys = ("distro", "release")
                kwargs.update({k: v for k, v in config.items() if k in keys})

                cls.create_container(repos=command_config.repos_path,
                                     packages=command_config.packages_path,
                                     **kwargs)

        def test_distro_folder_exists(self):
            """Check that distro folder exists for ."""
            if platform.system() == "Linux":
                root = get_dir_for_distro(self.container_dir,
                                          config)
                self.assertThat(os.path.join(self.container_dir, root),
                                DirExists())
            elif platform.system() == "Darwin":
                self.assertThat(os.path.join(self.container_dir, "bin"),
                                DirExists())

        def test_has_package_installed(self):
            """Check that our testing package got installed.

            If it did get installed, then it means that the repository
            was successfully added and the package was successfully installed
            using the native tool. That means that the proot "works".
            """
            format_kwargs = dict()

            if platform.system() == "Linux":
                root = get_dir_for_distro(self.container_dir,
                                          config)
                distro_arch = architecture.Alias.debian(kwargs["arch"])
                archlib = ARCHITECTURE_LIBDIR_MAPPINGS[distro_arch]
                format_kwargs["archlib"] = archlib
            else:
                root = self.container_dir

            # Match against a list of files. If none of the results are None,
            # then throw a list of mismatches.
            match_results = []
            for filename in test_files:
                path_to_file = os.path.join(root,
                                            filename.format(**format_kwargs))
                result = FileExists().match(path_to_file)
                if result:
                    match_results.append(result)

            if len(match_results) == len(test_files):
                raise Exception(repr(match_results))

    TemplateDistroTest.__name__ = test_name
    return TemplateDistroTest

_DistroPackage = namedtuple("_DistroPackage", "package files repo")
_DISTRO_INFO = {
    "Ubuntu": _DistroPackage(package="libaacs0",
                             repo=["{ubuntu} {release} universe"],
                             files=["usr/lib/{archlib}/libaacs.so.0"]),
    "Debian": _DistroPackage(package="libaio1",
                             repo=[],
                             files=["lib/libaio.so.1.0.1",
                                    "lib/{archlib}/libaio.so.1.0.1"]),
    "Fedora": _DistroPackage(package="libaio",
                             repo=[],
                             files=["lib/libaio.so.1.0.1",
                                    "lib64/libaio.so.1.0.1"]),
    "OSX": _DistroPackage(package="xz",
                          repo=[],
                          files=["bin/xz"]),
    "Windows": _DistroPackage(package="cmake.portable",
                              repo=[],
                              files=["lib/cmake.portable.3.4.3/"
                                     "tools/cmake-3.4.3-win32-x86/bin/"
                                     "cmake.exe"])
}


def get_distribution_tests():
    """Fetch distribution tests as dictionary."""
    tests = {}

    for config in available_distributions():
        config = config.copy()
        name_array = bytearray()
        for key in sorted(list(config.keys())):
            if key in ("info", "pkgsys", "url"):
                continue

            name_array += bytes(key[0].upper().encode() +
                                key[1:].encode() +
                                config[key][0].upper().encode() +
                                config[key][1:].encode())
        name = "Test{0}".format(name_array.decode("ascii"))

        distro = config["distro"]
        repositories_to_add = _DISTRO_INFO[distro].repo
        packages_to_install = [_DISTRO_INFO[distro].package]
        files_to_test_for = _DISTRO_INFO[distro].files
        kwargs = dict()

        try:
            kwargs["arch"] = Alias.universal(config["arch"])
        except KeyError:  # suppress(pointless-except)
            pass

        tests[name] = _create_distro_test(name,
                                          config,
                                          repositories_to_add,
                                          packages_to_install,
                                          files_to_test_for,
                                          **kwargs)

    return tests

for _name, _test in get_distribution_tests().items():
    exec("{0} = _test".format(_name))  # pylint:disable=W0122
    del _test



# /psqtraviscontainer/architecture.py
#
# Module which provides a helper for determining the names of various
# processor architectures on various distributions.
#
# See /LICENCE.md for Copyright information
"""Architecture handling."""

from collections import namedtuple

from six import with_metaclass

_ArchitectureType = namedtuple("_ArchitectureType",
                               "aliases debian universal qemu")

_X86_ARCHITECTURE = _ArchitectureType(aliases=["i386",
                                               "i486",
                                               "i586",
                                               "i686",
                                               "x86"],
                                      debian="i386",
                                      universal="x86",
                                      qemu="i386")
_X86_64_ARCHITECTURE = _ArchitectureType(aliases=["x86_64", "amd64"],
                                         debian="amd64",
                                         universal="x86_64",
                                         qemu="x86_64")
_ARM_HARD_FLOAT_ARCHITECTURE = _ArchitectureType(aliases=["arm",
                                                          "armel",
                                                          "armhf"],
                                                 debian="armhf",
                                                 universal="arm",
                                                 qemu="arm")
_POWERPC32_ARCHITECTURE = _ArchitectureType(aliases=["powerpc", "ppc"],
                                            debian="powerpc",
                                            universal="ppc",
                                            qemu="ppc")
_POWERPC64_ARCHITECTURE = _ArchitectureType(aliases=["ppc64el", "ppc64"],
                                            debian="ppc64el",
                                            universal="ppc64",
                                            qemu="ppc64")


class _AliasMetaclass(type):
    """A metaclass which provides an operator to convert arch strings."""

    @classmethod
    def __getitem__(cls,  # pylint:disable=bad-mcs-classmethod-argument
                    lookup):
        """Operator overload for [].

        If a special architecture for different platforms is not found, return
        a generic one which just has this architecture name
        """
        del cls

        overloaded_architectures = [_X86_ARCHITECTURE,
                                    _X86_64_ARCHITECTURE,
                                    _ARM_HARD_FLOAT_ARCHITECTURE,
                                    _POWERPC32_ARCHITECTURE,
                                    _POWERPC64_ARCHITECTURE]
        for arch in overloaded_architectures:
            if lookup in arch.aliases:
                return arch

        return _ArchitectureType(aliases=[lookup],
                                 debian=lookup,
                                 universal=lookup,
                                 qemu=lookup)


class Alias(with_metaclass(_AliasMetaclass, object)):
    """Implementation of _AliasMetaclass.

    Provides convenience methods to convert architecture strings
    between platforms.
    """

    @classmethod
    def debian(cls, lookup):
        """Convert to debian."""
        return cls[lookup].debian

    @classmethod
    def qemu(cls, lookup):
        """Convert to qemu."""
        return cls[lookup].qemu

    @classmethod
    def universal(cls, lookup):
        """Convert to universal."""
        return cls[lookup].universal



# /psqtraviscontainer/common_options.py
#
# Options common to both both commands.
#
# See /LICENCE.md for Copyright information
"""Options common to both both commands."""

import platform

import configargparse

from psqtraviscontainer import architecture
from psqtraviscontainer import distro


def get_parser(action):
    """Get a parser with options common to both commands."""
    # Iterate over the available_distributions and get a list of available
    # distributions and architectures for the --distro and --arch arguments
    architectures = set()
    distributions = set()

    for config in distro.available_distributions():
        if "distro" in config:
            distributions.add(config["distro"])
        if "arch" in config:
            architectures.add(architecture.Alias.universal(config["arch"]))

    description = """{0} a CI container""".format(action)
    parser = configargparse.ArgumentParser(description=description)

    current_arch = architecture.Alias.universal(platform.machine())

    parser.add_argument("containerdir",
                        metavar=("CONTAINER_DIRECTORY"),
                        help="""Directory to place container in""",
                        type=str)
    parser.add_argument("--distro",
                        type=str,
                        help="""Distribution name to create container of""",
                        choices=distributions,
                        env_var="CONTAINER_DISTRO")
    parser.add_argument("--release",
                        type=str,
                        help="""Distribution release to create container of""",
                        env_var="CONTAINER_RELEASE")
    parser.add_argument("--arch",
                        type=str,
                        help=("""Architecture (all architectures other """
                              """than the system architecture will be """
                              """emulated with qemu)"""),
                        default=current_arch,
                        choices=architectures,
                        env_var="CONTAINER_ARCH")

    return parser



# /psqtraviscontainer/create.py
#
# Module which handles the creation of proot in which APT packages
# can be readily installed
#
# See /LICENCE.md for Copyright information
"""Module which handles the creation of proot.

The proot for a distribution is a special directory entered with the proot
command, which behaves like a chroot, except that no root access is required
in order to create the jail. Commands running in the proot will have filesystem
requests redirected to the proot and believe that they are running as uid 0
"""

from __future__ import unicode_literals

import os

from clint.textui import colored

from psqtraviscontainer import common_options
from psqtraviscontainer import distro
from psqtraviscontainer import printer

from psqtraviscontainer.architecture import Alias


def _print_distribution_details(details):
    """Print distribution details."""
    def _y_v(value):
        """Print value in distribution details."""
        return colored.yellow(value)

    # Maps keys in configuration to a pretty-printable name.
    distro_pretty_print_map = {
        "distro": lambda v: """Distribution Name: """ + _y_v(v),
        "release": lambda v: """Release: """ + _y_v(v),
        "arch": lambda v: """Architecture: """ + _y_v(Alias.universal(v)),
        "pkgsys": lambda v: """Package System: """ + _y_v(v.__name__),
    }

    output = bytearray()
    output += ("\n" +
               colored.white("""Configured Distribution:""", bold=True) +
               "\n").encode()

    for key, value in details.items():
        if key in distro_pretty_print_map:
            line = " - {0}\n".format(distro_pretty_print_map[key](value))
            output += line.encode()

    output += "\n".encode()

    printer.unicode_safe(output.decode("utf-8"))


def _parse_arguments(arguments=None):
    """Return a parser context result."""
    parser = common_options.get_parser("Create")
    parser.add_argument("--repositories",
                        type=str,
                        help="""A file containing a list of repositories to """
                             """add before installing packages. Special """
                             """keywords will control the operation of this """
                             """file: \n"""
                             """{release}: The distribution release (eg, """
                             """precise)\n"""
                             """{ubuntu}: Ubuntu archive URL\n"""
                             """{launchpad}: Launchpad PPA URL header (eg,"""
                             """ppa.launchpad.net)\n""",
                        default=None)
    parser.add_argument("--packages",
                        type=str,
                        help="""A file containing a list of packages """
                             """to install""",
                        default=None)

    return parser.parse_args(arguments)


def main(arguments=None):
    """Parse arguments and set up proot.

    Parse arguments, fetches initial proot distribution and downloads
    and sets up our proot.
    """
    result = _parse_arguments(arguments=arguments)
    container_dir = os.path.realpath(result.containerdir)

    selected_distro = distro.lookup(vars(result))

    _print_distribution_details(selected_distro)

    # Now set up packages in the distribution. If more packages need
    # to be installed or the installed packages need to be updated then
    # the build cache should be cleared.
    with selected_distro["info"].create_func(container_dir,
                                             selected_distro) as container:
        container.install_packages(result.repositories, result.packages)

    relative_containerdir = os.path.relpath(result.containerdir)
    msg = """\N{check mark} Container has been set up in {0}\n"""
    printer.unicode_safe(colored.green(msg.format(relative_containerdir),
                                       bold=True))

if __name__ == "__main__":
    main()



# /psqtraviscontainer/download.py
#
# Module with utilities for downloading files.
#
# See /LICENCE.md for Copyright information
"""Module with utilities for downloading files."""

import os

import sys

from clint.textui import colored, progress

import requests


def download_file(url, filename=None):
    """Download the file at url and store it at filename."""
    basename = os.path.basename(filename or url)
    msg = """Downloading {dest} (from {source})""".format(source=url,
                                                          dest=basename)
    sys.stdout.write(str(colored.blue(msg, bold=True)))
    sys.stdout.write("\n")
    request = requests.get(url, stream=True)
    length = int(request.headers.get("content-length"))
    with open(filename or os.path.basename(url), "wb") as downloaded_file:
        chunk_size = 1024
        total = length / chunk_size + 1
        for chunk in progress.bar(request.iter_content(chunk_size=chunk_size),
                                  expected_size=total,
                                  label=basename):
            downloaded_file.write(chunk)
            downloaded_file.flush()

    return os.path.join(os.getcwd(), downloaded_file.name)


class TemporarilyDownloadedFile(object):  # pylint:disable=R0903
    """An enter/exit class representing a temporarily downloaded file.

    The file will be downloaded on enter and erased once the scope has
    been exited.
    """

    def __init__(self, url, filename=None):
        """Initialize the url and path to download file to."""
        super(TemporarilyDownloadedFile, self).__init__()
        self._url = url
        self._path = download_file(self._url, filename)

    def __enter__(self):
        """Run file download."""
        return self

    def __exit__(self, exc_type, value, traceback):
        """Remove the temporarily downloaded file."""
        del exc_type
        del traceback
        del value

        os.remove(self._path)
        self._path = None

    def path(self):
        """Get temporarily downloaded file path."""
        return self._path



# /setup.py
#
# Installation and setup script for psqtraviscontainer
#
# See /LICENCE.md for Copyright information
"""Installation and setup script for psqtraviscontainer."""

import platform

from setuptools import find_packages, setup

INSTALL_EXTRAS = []

if platform.system() != "Windows":
    INSTALL_EXTRAS.extend([
        "python-debian"
    ])

setup(name="polysquare-travis-container",
      version="0.0.21",
      description="""Polysquare Travis-CI Container Root""",
      long_description_markdown_filename="README.md",
      author="Sam Spilsbury",
      author_email="smspillaz@gmail.com",
      classifiers=["Development Status :: 3 - Alpha",
                   "Intended Audience :: Developers",
                   "Topic :: Software Development :: Build Tools",
                   "License :: OSI Approved :: MIT License",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3.3",
                   "Programming Language :: Python :: 3.4"],
      url="http://github.com/polysquare/polysquare-travis-container",
      license="MIT",
      keywords="development travis",
      packages=find_packages(exclude=["test"]),
      install_requires=["clint",
                        "configargparse",
                        "parse-shebang>=0.0.3",
                        "requests",
                        "six",
                        "shutilwhich",
                        "tempdir"] + INSTALL_EXTRAS,
      extras_require={
          "green": [
              "testtools",
              "shutilwhich",
              "nose",
              "nose-parameterized>=0.0.5",
              "mock",
              "setuptools-green>=0.0.14",
              "tempdir"
          ],
          "polysquarelint": [
              "polysquare-setuptools-lint>=0.0.25"
          ],
          "upload": [
              "setuptools-markdown"
          ]
      },
      entry_points={
          "console_scripts": [
              "psq-travis-container-create=psqtraviscontainer.create:main",
              "psq-travis-container-exec=psqtraviscontainer.use:main",
              "psq-travis-container-get-root=psqtraviscontainer.rootdir:main"
          ]
      },
      test_suite="nose.collector",
      zip_safe=True,
      include_package_data=True)



# /psqtraviscontainer/util.py
#
# Utility functions for working with files.
#
# See /LICENCE.md for Copyright information
"""Utility functions for working with files."""

import os


def check_if_exists(entity):
    """Raise RuntimeError if entity does not exist."""
    if not os.path.exists(entity):
        raise RuntimeError("""A required entity {0} does not exist\n"""
                           """Try running psq-travis-container-create """
                           """first before using psq-travis-container-use."""
                           """""".format(entity))



# /psqtraviscontainer/directory.py
#
# Utilities for safe directory navigation.
#
# See /LICENCE.md for Copyright information
"""Utilities for safe directory navigation."""

import errno

import os


class Navigation(object):  # pylint:disable=R0903
    """Context manager to enter and exit directories."""

    def __init__(self, path):
        """Initialize the path we want to change to."""
        super(Navigation, self).__init__()
        self._path = path
        self._current_dir = None

    def __enter__(self):
        """Upon entry, attempt to create the directory and then enter it."""
        try:
            os.makedirs(self._path)
        except OSError as err:
            if err.errno != errno.EEXIST:  # suppress(PYC90)
                raise err

        self._current_dir = os.getcwd()
        os.chdir(self._path)

        return self._path

    def __exit__(self, exc_type, value, traceback):
        """Pop directory on exiting with statement."""
        del exc_type
        del traceback
        del value

        os.chdir(self._current_dir)



# /test/testutil.py
#
# Monkey-patch functions for tests to cache downloaded files. This helps
# to speed up test execution.
#
# See /LICENCE.md for Copyright information
"""Monkey-patch functions for tests to cache downloaded files."""

import errno

import hashlib

import os

import shutil

import sys

from clint.textui import colored

from psqtraviscontainer.download import download_file as download_file_original


def download_file_cached(url, filename=None):
    """Check if we've got a cached version of url, otherwise download it."""
    cache_dir = os.environ.get("_POLYSQUARE_TRAVIS_CONTAINER_TEST_CACHE_DIR",
                               None)
    if cache_dir:
        try:
            os.makedirs(cache_dir)
        except OSError as error:
            if error.errno != errno.EEXIST:  # suppress(PYC90)
                raise error

        dest_filename = os.path.realpath(filename or os.path.basename(url))
        hashed = os.path.join(cache_dir,
                              hashlib.md5(url.encode("utf-8")).hexdigest())

        if os.path.exists(hashed):
            msg = """Downloading {0} [found in cache]\n""".format(url)
            sys.stdout.write(str(colored.blue(msg, bold=True)))
            shutil.copyfile(hashed, dest_filename)
        else:
            # Grab the url and then store the downloaded file in the cache.
            # We trust download_file_original to give us the
            # right dest_filename, hence the reason why we overwrite it here.
            dest_filename = download_file_original(url, filename)
            shutil.copyfile(dest_filename, hashed)
    else:
        dest_filename = download_file_original(url, filename)

    return dest_filename



# /psqtraviscontainer/use.py
#
# Module which handles the running of scripts and commands inside of a proot
#
# See /LICENCE.md for Copyright information
"""Module which handles the running of scripts inside of a proot."""

import os

import sys

from psqtraviscontainer import common_options
from psqtraviscontainer import distro


def _parse_arguments(arguments=None):
    """Return a parser context result."""
    parser = common_options.get_parser("Use")
    parser.add_argument("--show-output",
                        action="store_true",
                        help="""Show output of commands once they've run.""")
    return parser.parse_args(arguments)


def main(arguments=None):
    """Select a distro in the container root and runs a command in it."""
    arguments = (arguments or sys.argv[1:])

    try:
        two_dashes_argument = arguments.index("--")
    except ValueError:
        sys.stdout.write("""Command line must specify command to """
                         """run with two dashes\n""")
        sys.exit(1)

    parseable_arguments = arguments[:two_dashes_argument]
    command = arguments[two_dashes_argument + 1:]

    argparse_result = _parse_arguments(arguments=parseable_arguments)

    container_dir = os.path.realpath(argparse_result.containerdir)
    selected_distro = distro.lookup(vars(argparse_result))
    with selected_distro["info"].get_func(container_dir,
                                          selected_distro) as container:
        if argparse_result.show_output:
            execute_kwargs = {
                "stderr": None,
                "stdout": None
            }
        else:
            execute_kwargs = dict()

        result = container.execute(command, **execute_kwargs)[0]

    return result



# /psqtraviscontainer/distro.py
#
# Module in which all configurations for various distributions are stored.
#
# See /LICENCE.md for Copyright information
"""Various distribution configurations are stored here."""

import itertools

from collections import namedtuple


DistroConfig = dict
DistroInfo = namedtuple("DistroInfo",
                        "create_func "
                        "get_func "
                        "match_func "
                        "enumerate_func "
                        "kwargs")


def _distribution_information():
    """Return generator of DistroInfo."""
    from psqtraviscontainer import linux_container
    from psqtraviscontainer import osx_container
    from psqtraviscontainer import windows_container

    return itertools.chain(linux_container.DISTRIBUTIONS,
                           osx_container.DISTRIBUTIONS,
                           windows_container.DISTRIBUTIONS)


def available_distributions():
    """Return list of available distributions."""
    for info in _distribution_information():
        for config in info.enumerate_func(info):
            config["info"] = info
            config = config.copy()
            yield config


def lookup(arguments):
    """Look up DistroConfig by matching against its name and arguments."""
    matched_distribution = None

    for distribution in _distribution_information():
        matched_distribution = distribution.match_func(distribution,
                                                       arguments)
        if matched_distribution:
            matched_distribution["info"] = distribution
            return matched_distribution

    raise RuntimeError("""Couldn't find matching distribution """
                       """({0})""".format(repr(arguments)))
